!********************************************************************************
!*   Copyright 1999 -- 2002 C. Kind, M. Reiher, J. Neugebauer, and B. A. Hess,
!*   Chair of Theoretical Chemistry, Friedrich-Alexander-University 
!*   Erlangen-Nuremberg, Egerlandstr. 3, D-91058 Erlangen, Germany.
!*
!*
!*   This file is part of the package SNF.
!*
!*   SNF is free software; you can redistribute it and/or modify
!*   it under the terms of the GNU General Public License as published by
!*   the Free Software Foundation; either version 2 of the License, or
!*   (at your option) any later version.
!*
!*   This program is distributed in the hope that it will be useful,
!*   but WITHOUT ANY WARRANTY; without even the implied warranty of
!*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!*   GNU General Public License for more details.
!*
!*   You should have received a copy of the GNU General Public License
!*   along with this program; if not, write to the Free Software
!*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!*
!********************************************************************************

program snf

  use f90_kind

! ===========================================================================
! The program SNF calculates vibrational frequencies and intensities for
! infrared, Raman, and vibrational Raman optical activity spectra.
!
! The calculations are performed by numerical differentiation of raw data
! (gradients of the electronic energy and property tensors) stored in the
! file "restart". These data are provided by the data collector SNFDC,
! which is provided together with snf in this package. SNFDC is docked to
! several electronic structure programs, which actually calculate the raw data
! for SNF. Currently, the following programs are interfaced to SNF:
!  1) Turbomole
!  2) Dalton
!  3) ADF
!  4) Gaussian
!  5) Molpro (currently only for MCSCF)
!
! SNF uses the full point group symmetry of the molecule for constructing,
! analyzing and diagonalizing the Hessian.
!
! history:
!
! - written by Carsten Kind, Markus Reiher and Johannes Neugebauer, 
!   Uni Erlangen 1999-2003
! - extended by M. Reiher, Uni Erlangen, June 2003
! - modified by M. Reiher, Uni Bonn, October 2004
! - modified by C. Herrmann, Uni Bonn/Jena/ETH Zuerich, 2004-2006
! - modified by S. Luber, ETH Zuerich, June/October 2006
! ===========================================================================

  use snf_spectrum
  use snf_molecule
  use snf_turbomole
  use snf_char_table
  use snf_representation
  use snf_projection
  use snf_character
  use snf_hessian
  use snf_restart
  use snf_error
  use snf_xmol
  use snf_rdalmol
  use snf_adfmol
  use snf_gaussmol
  use snf_molpmol
  use snf_control
  use snf_dipol
  use snf_polar
  use snf_vroa
  use snf_irred_dist
  use snf_irred_atoms
  use snf_thermo
  use snf_tools
  use snf_norm_modes
  use snf_gaussian
  use snf_mopac
  use snf_cioslowski
  use snf_gnuplot
  use snf_param
  use snf_symmat
  use snf_dummy

  implicit none

! --------------------------------------------------------------------------------------------------------
! variables ...
! --------------------------------------------------------------------------------------------------------
  type(type_ctrl)              :: ctrl,ctrl_restart  ! essential control flags and parameters
  type(type_atom), pointer     :: molecule(:)        ! geometry atoms and atomic masses of the molecule  
  type(type_atom), allocatable :: mol(:)
  type(type_molecule)          :: molekel            ! information about the whole molecule
  real(double), pointer        :: gradients(:,:,:)   ! energy gradients of the negative distortion (nvar,nvar)
  real(double), pointer        :: dipols(:,:,:)      ! dipol moments of the negative distortion (nvar,3)
  real(double), pointer        :: pols(:,:,:,:)      ! polarizabilities pols(1..8,nvar=3*n_atoms,
                                                     ! no_steps=no_distortions,no_frequencies(dynamic pols))
  real(double), pointer        :: gtenlao(:,:,:,:),  &
                                  gtenao(:,:,:,:),   &
                                  aten(:,:,:,:)      ! VROA tensors
  real(double), allocatable    :: distortion(:)      ! generalized distortion in the cartesian basis

  real(double), pointer        :: c2p(:,:)           ! transformation matrix from cartesian 
  real(double), pointer        :: hessian_xyz(:,:)   ! hessian in the cartesian basis (nvar,nvar)
  real(double), allocatable    :: hessian_m(:,:),    &! hessian in the mass weighted cartesian basis (nvar,nvar)
                                  hessian_p(:,:),    &! hessian in the symmetryadapted basis of (irrep,irrep)
                                  hessian_q(:),      &! hessian in the basis of normal modes (irrep)
                                  norm_modes(:,:)     ! transformation matrix from symmetryadapted
                                                      ! basis to normal coordinates (irrep,irrep)
                                                      ! transformation matrix from cartesian basis 
                                                      ! to normal coordinates  (nvar,irrep)
  real(double), pointer        :: &
                                  dipol_dxyz(:,:),    &! deriv. of dipol moments in cartesian basis (nvar,3)
                                  dipol_dq(:,:),      &! deriv. of dipol moments in normal coordinates (irrep,3)
                                  polar_dxyz(:,:,:),  &! deriv. of polarizability tensor (first dimension: 6) in cartesian basis (nvar,?)
                                  polar_dq(:,:,:),    &! deriv. of polarizability tensor in normal coordinates (irrep,?)
                                  polar9_dq(:,:,:),   &! deriv. of polarizability tensor (first dimension: 9) in normal coordinate basis (nvar,?)
                                  gtenlao_dxyz(:,:,:),&! deriv. of G/LAO tensor in cartesian basis (nvar,?)
                                  gtenlao_dq(:,:,:),  &! deriv. of G/LAO tensor in normal coordinates (irrep,?)
                                  gtenao_dxyz(:,:,:), &! deriv. of G/AO tensor in cartesian basis (nvar,?)
                                  gtenao_dq(:,:,:),   &! deriv. of G/AO tensor in normal coordinates (irrep,?)
                                  aten_dxyz(:,:,:),   &! deriv. of A tensor in cartesian basis (nvar,?)
                                  aten_dq(:,:,:)       ! deriv. of A tensor in normal coordinates (irrep,?)

  real(double), pointer        :: &
                                  representation(:,:,:), &! representations of symmetry operations s -->(s,:,:)
                                  proj_op(:,:,:),        &! projectors on irreps i-->(i,:,:)
                                  character(:)            ! characters of the representations

  integer, pointer             :: irreps(:)            ! vector with number of vectors per irrep
  type(type_char_table)        :: char_table           ! character table
  type(type_peak), pointer     :: spectrum(:),        &! spectral information with frequencies, modes and intensities 
                                  global_spectrum(:), &! spectral information with frequencies, modes and intensities 
                                  sum_spectrum(:)      ! spectral information, summed over degenerate modes
  integer                      :: irrep,              &! index of irrep
                                  irrep_dim            ! number of vectors in actual irrep, therefore 
                                                       ! dimension parameter
  integer, pointer             :: &
                                  istepfl(:),         &! points to calculate(=6*nirred_atoms)
                                  atom_index(:),      &! index of irreducible atom (1:natoms)
                                  symm_index(:),      &! symmetry op. to optain irred. atom (1:natoms)
                                  selatom(:),         &! atom-numbers which special isotopes are selected for
                                  irred_atom_index(:)  ! index list with irreducible atoms (1:nirred_atoms)

  integer, pointer             :: dist_index(:),      &! index of irreducible dist. 
                                  dsymm_index(:),     &! symmetry op. to optain irred. dist. 
                                  irred_dist_index(:)  ! index list with irreducible dist. 

  character(len=8), pointer    :: nvmach(:)            ! list of all used machines 
  character(len=132)           :: errbuff              ! buffer for error messages
  character(len=132)           :: tbuf                 ! buffer for time messages

  real(double), allocatable    :: eig_proj_op(:)
  real(double)                 :: zeropoint            ! zero point energy
  real(double), pointer        :: massatom(:)          ! masses for special isotopes

  integer                      :: zeros,t1,t2,ltbuf
  real(double)                 :: dumlaserf(1),ddum
 
  real(double)                 :: mol_mass
  real(double), pointer        :: freqs(:)
  real(double), pointer        :: cioslowski(:)        ! cioslowski charge decomposition
  character(len=7), pointer    :: prod_table(:,:)

  integer                      :: i,ii,j,k,nirred_atoms,nvar,offset,nfreq,nirred_dist
  integer                      :: number_projected_modes,out
  logical                      :: new_restart,fdum,static,haflag,rflag,vroaflag
  integer                      :: lwork,info,ilaenv,status
  real(double),allocatable     :: work(:),tmpmat(:,:)
  real(double),allocatable     :: dip_contr(:,:,:),pol_contr(:,:,:)
  integer                      :: csv,natoms
  real(double),allocatable     :: normmodes(:,:)       ! intensities-only mode: extra field for normal modes 

! --------------------------------------------------------------------------------------------------------
! end of variables
! --------------------------------------------------------------------------------------------------------

  call secs(t1)
  csv=77
  out=44
  open(unit=out,file='snf.out',status='unknown')
  call print_header(out)
  call last_mod(out)
  call snf_cite_as(out)
  nullify(molecule,gradients,dipols,pols,gtenlao,gtenao,aten,&
       &  hessian_xyz,c2p,                                   &
       &  dipol_dxyz,dipol_dq,polar_dxyz,polar_dq,           &
       &  gtenlao_dxyz,gtenlao_dq,gtenao_dxyz,gtenao_dq,     &
       &  aten_dxyz,aten_dq,polar9_dq,                       &
       &  representation,proj_op,character,irreps,           &
       &  cioslowski,nvmach,                                 &
       &  irred_atom_index,symm_index,atom_index,istepfl,    &
       &  global_spectrum,spectrum,sum_spectrum)
  !----------------------------------------------------------------
  ! read coord file, write data to "molecule"
  ! determine symmetry and distortion from control file
  ! read gradients and dipol moments, and polarizabilities from restart file
  !----------------------------------------------------------------
  call init_ctrl(ctrl)
  call testfile('snf_control')
  call testfile('restart')
  call read_control(ctrl,selatom,massatom)

  if(ctrl%progno.eq.1) then
!----------
! TURBOMOLE
!----------
    call testfile('control')
    call testfile('coord')
  else if(ctrl%progno.eq.2) then
!-------
! DALTON
!-------
    call testfile('molecule.dal')
    call testfile('molecule.mol')
  else if(ctrl%progno.eq.3) then
!-------
! Gaussian
!-------
     call testfile('gaussian.com')
  else if(ctrl%progno.eq.4) then
!-------
! ADF
!-------
     call testfile('adf.in')
  else if(ctrl%progno.eq.5) then 
!-----
! MOLPRO
!------
     call testfile('molpro.in')
  end if
  if(associated(ctrl%tm_ctrl)) then
!     ctrl%calcno=ctrl%tm_ctrl%calcno
     call read_tm_coord(ctrl%tm_ctrl%coord,molecule)
  else if(associated(ctrl%dal_ctrl)) then
     call read_dal_coord(ctrl,molecule)
  else if(associated(ctrl%gauss_ctrl)) then
     call read_gauss_coord(ctrl,molecule)
  else if(associated(ctrl%adf_ctrl)) then
     call read_adf_coord(ctrl,molecule)
  else if(associated(ctrl%molp_ctrl)) then
     call read_molp_coord(ctrl,molecule)
  endif
  if(associated(selatom)) then
     call setisomass(molecule,selatom,massatom)
     deallocate(selatom,massatom)
  end if
  fdum=.false.
  do i=1,ubound(molecule,1)
    if(molecule(i)%element.eq.'Q') fdum=.true.
  end do
  ! calculate molecule mass and moments of inertia
  nvar=3*ubound(molecule,1)
  if(.not.is_point_grp(molecule,1.D-3,atom_equal,ctrl%symmetry)) then
     errbuff=''
     write(errbuff,'(2a)') 'Molecule does not fit into symmetry:',ctrl%symmetry
     call serror('snf',2,errbuff,out)
  endif
  if(ctrl%isosym.ne.'def') then
    if(.not.is_point_grp(molecule,1.D-3,isotope_equal,ctrl%isosym)) then
      errbuff=''
      write(errbuff,'(2a)') 'Molecule does not fit into symmetry:',ctrl%isosym
      call serror('snf',2,errbuff,out)
    endif
  endif
  ctrl%nvar=nvar

  call print_ctrl(ctrl,unit=out)
  if(any(molecule(:)%changem)) call print_isomass(molecule,unit=out)
  call print_molecule(molecule,unit=out)

  if(associated(ctrl%dynfreq)) then
    nfreq=ubound(ctrl%dynfreq,1)
  else
    nfreq=1
  end if
  allocate(ctrl_restart%dynfreq(nfreq))
  rflag=ctrl%ramanew.ge.1
  vroaflag=ctrl%vroaew.ge.1
! raman and vroa modes should never be possible at the same time:
  if (rflag.and.vroaflag) stop " MIRACLE "
  call read_restart(ctrl%restart,ctrl_restart,&
       &   gradients,dipols,pols,gtenlao,gtenao,aten,&
       &    istepfl,nvmach,status,rflag,&
       &   vroaflag,unit=out)
  if(status.eq.-9) then
    call serror('snf',2,'Restart file does not exist.',out)
  else if(status.eq. -1) then
    call serror('snf',2,'Some calculations are not finished.',out)
  else if(status.eq. 0) then
    call serror('snf',2,'Not all calculations have been done yet.',out)
  else if(status.eq. -2) then
    call serror('snf',2,'Not all data necessary for frequency analysis##&
     &available in file restart.',out)
  endif
  call check_ctrl(ctrl_restart,ctrl,out)

  call destroy_control(ctrl_restart)
  if(fdum) then
    call serror('snf',-1,'Dummy atoms are neglected in all further &
     &evaluations',out)
    call rm_dummies(molecule,gradients,dipols,pols,gtenlao,gtenao,aten,nvmach,&
     & istepfl,2*ctrl%numderiv,rflag,vroaflag,nfreq)
    nvar=3*ubound(molecule,1)
    ctrl%nvar=nvar
  end if
!-------------------------------------------------
! intensities-only-mode:
! allocate fields an read control for intonly mode
! skip sections which are only needed
! for calculation of normal modes and property derivatives
! along Cartesian coordinates ...
!-------------------------------------------------
  if(ctrl_restart%nummodes.ne.0) then
    write(*,*)
    write(*,*) ' ***************************************'
    write(*,*) ' Running SNF in intensities-only-mode...'
    write(*,*) ' ***************************************'
    write(*,*)
    write(out,*)
    write(out,*) ' ***************************************'
    write(out,*) ' Running SNF in intensities-only-mode...'
    write(out,*) ' ***************************************'
    write(out,*)
! allocate fields for normal modes and freqs. read in
    if(.not.associated(ctrl%nmodein)) allocate(ctrl%nmodein(nvar, &
   &nvar))
    allocate(ctrl%freqin(nvar*3))
    allocate(ctrl%selectmode(nvar*3))
    call read_control2(ctrl,ubound(molecule,1))
    goto 1000
  end if

  !----------------------------------------------------------------
  ! one distortion value for all cartesian coordinates
  ! note: generalized distortions may be necessary in order to 
  !       perform distortions along the normal coordinates
  !----------------------------------------------------------------
  allocate(distortion(nvar))
  distortion(:)=ctrl%cstep  
  !----------------------------------------------------------------
  ! determine total number of translational and rotational modes
  !----------------------------------------------------------------  
  allocate(mol(ubound(molecule,1)))
  mol=molecule
  haflag=.false.
  call hauptax_trans(mol,molekel,haflag)
  deallocate(mol)
  zeros=0
  do i=1,3
     if(molekel%p_inertia(i).lt. 1.d0) zeros=zeros+1
  enddo

  number_projected_modes=6-zeros
  !----------------------------------------------------------------
  ! construct symmetry information
  !    character table
  !    representations
  !    projection operators
  !    characters and irreducible representations
  !    list of irreducible atoms with symmetry operations
  !----------------------------------------------------------------
  call create_char_table(char_table,ctrl%symmetry)
  call check_geometry(molecule,char_table,atom_equal,unit=out)
  call create_irred_atoms(molecule,char_table,atom_index,symm_index, &
       &                  irred_atom_index,nirred_atoms)
  if(ctrl%print_irred_at) &
       & call print_irred_atoms(atom_index,symm_index,&
       &         irred_atom_index,nirred_atoms,unit=out)
  !----------------------------------------------------------------
  ! create hessian matrix from gradient and perform mass weighting
  ! calculate cartesian derivative dipol moments and print them out
  !----------------------------------------------------------------
  if(ctrl%maxsym) call create_irred_dist(molecule,char_table,dist_index, &
       &  dsymm_index,irred_dist_index,nirred_dist,nirred_atoms,&
       &  irred_atom_index,2*ctrl%numderiv,distortion,&
       &  ctrl%print_irred_dist,unit=out)
!MR Vorsicht: In der folgenden Routine stuerzt snf bei c1 symmetrie ab!!!
!MR           und zwar mit folgender Fehlermeldung: ERROR calling the intrinsic subprogram MATMUL:
!MR                                                 At least one of the array arguments should have rank = 2
!MR dies tritt nur mit dem ifc Compiler auf!! BSP: nf_3.1.0/example/tst
!MR dasselbe Problem tritt fuer snf_define in snf_irred_atoms (beim ifc) auf!
  call create_representation_for_vib(representation,molecule,char_table)
!MR
  call create_character(representation,character)
  call create_irreps(character,irreps,char_table)
  call create_proj_op(representation,char_table,proj_op)
  if(ctrl%print_symmetry) then
     call print_char_table(char_table,unit=out)
     call create_product_table(char_table,prod_table)
     call print_product_table(char_table,prod_table,unit=out)
     deallocate(prod_table)
     write(out,*)
     write(out,*)'Symmetry elements'
     write(out,*)'-----------------'
     do i=1,char_table%nsymel
        call printmat(symel_mat(char_table%symel(i)),text=char_table%symel(i),unit=out,fmt='F10.5')
     enddo
     ! call print_representation(representation,char_table,unit=out)
     ! call print_proj_op(proj_op,char_table,unit=out)
     call print_character(character,char_table,unit=out)
     call print_irreps(irreps,char_table,unit=out)
  endif
  if(ctrl%maxsym) call create_red_grad(gradients,dist_index, &
     & dsymm_index,nirred_atoms,&
     & irred_atom_index,2*ctrl%numderiv,representation)
  if(ctrl%dtype.eq.'fward'.or.ctrl%dtype.eq.'bward') then
    errbuff=''
    write(errbuff,'(a)') 'Type of derivatives used for Hessian: ' &
   & //ctrl%dtype//'##DO NOT USE THIS TYPE OF DERIVATIVES IN##PRODUCTION CALCULATIONS'
    call serror('snf',0,errbuff,out)
    call serror('snf',0,errbuff,6)
    call create_hessian_xyz_fb(hessian_xyz,gradients,distortion,ctrl%dtype)
  else
    call create_hessian_xyz(hessian_xyz,gradients,distortion)
  end if
  call expand_hessian(hessian_xyz,representation,symm_index)
  call check_hessian(hessian_xyz,ctrl,symm_index,istepfl,new_restart,unit=out)
  call symmetrize_hessian(hessian_xyz,representation)
  if(new_restart) then
     if(fdum) then
       call serror('snf',-1,'SNF is not yet able to write new restart &
         &files##if dummies are present. The restart file has to be &
         &changed manually.',unit=out)
     else
       call backup_restart(ctrl%restart,unit=out)
       call write_restart(gradients,dipols,pols,gtenlao,gtenao,aten, &
                         &istepfl,nvmach,ctrl,unit=out)
     end if
  endif
  if(associated(nvmach))  deallocate(nvmach)
  if(associated(istepfl)) deallocate(istepfl)
  if(ctrl%print_hessian) &
       & call printmat(hessian_xyz,text='cartesian hessian',unit=out) 
  call balance_hessian(hessian_xyz)
  !call print_hessian(hessian_xyz,'nf_hessian.res')
  allocate(hessian_m(nvar,nvar))
  call hessian_mass_weighting(hessian_xyz,hessian_m,molecule)
  !----------------------------------------------------------------
  ! for dtype.ne.'centr' : do not calculate intensities
  !----------------------------------------------------------------
  if(ctrl%dtype.ne.'centr') goto 100
  !----------------------------------------------------------------
  ! calculate cartesian derivative dipol moments and 
  ! polarizabilities (if requested) and print them out 
  !----------------------------------------------------------------
  if(ctrl%maxsym) call create_red_dip(dipols,dist_index, &
     & dsymm_index,nirred_atoms,&
     & irred_atom_index,2*ctrl%numderiv,char_table)
  call create_dipol_dxyz(dipol_dxyz,dipols,distortion)
  call expand_dipol(dipol_dxyz,representation,symm_index,char_table)
  call symm_dipol(dipol_dxyz,representation,char_table)
  if(ctrl%print_dip) &
       & call printmat(dipol_dxyz,text='gradient of dipole moments',unit=out)
  if (ctrl%ramanew.ge.1.and.(.not.associated(ctrl%dynfreq))) then
     allocate(polar_dxyz(ubound(pols,1),ubound(pols,2),1))
     if(ctrl%maxsym) call create_red_pol(pols(:,:,:,1),dist_index, &
       & dsymm_index,nirred_atoms,&
       & irred_atom_index,2*ctrl%numderiv,char_table)
     call create_polar_dxyz(polar_dxyz(:,:,1),pols(:,:,:,1),distortion)
     call expand_polar(polar_dxyz(:,:,1),representation,symm_index,char_table)
     call symm_polar(polar_dxyz(:,:,1),representation,char_table)
     if(ctrl%print_pol) then
       call printmat(polar_dxyz(:,:,1),text='gradient of polarizability tensor',&
          & unit=out)
      end if
  else if(ctrl%ramanew.ge.1.and.associated(ctrl%dynfreq)) then
     allocate(polar_dxyz(ubound(pols,1),ubound(pols,2),ubound(ctrl%dynfreq,1)))
     do i=1,ubound(ctrl%dynfreq,1)
       if(ctrl%maxsym) call create_red_pol(pols(:,:,:,i),dist_index, &
          & dsymm_index,nirred_atoms,&
          & irred_atom_index,2*ctrl%numderiv,char_table)
       call create_polar_dxyz(polar_dxyz(:,:,i),pols(:,:,:,i),distortion)
       call expand_polar(polar_dxyz(:,:,i),representation,symm_index,char_table)
       call symm_polar(polar_dxyz(:,:,i),representation,char_table)
       if(ctrl%print_pol) then
         write(out,*)
         write(out,*)
         if(abs(ctrl%dynfreq(i)).le.1.0d-6) then
           write(out,'(a)') ' Dynamic polarizability derivatives for &
            &laser wavelength:  infinity'
         else
           write(out,'(a,f16.5)') ' Dynamic polarizability derivatives for &
            &laser wavelength: ',1.0d0/ctrl%dynfreq(i)*nm2freq
         end if
         write(out,*)
         call printmat(polar_dxyz(:,:,i),text='gradient of polarizability tensor',&
          & unit=out)
       end if 
     end do
  end if
!#####################################################################################
! allocate arrays for VROA calculations and calculate Cartesian gradients
!
  if (ctrl%vroaew.ge.1.and.(.not.associated(ctrl%dynfreq))) then
! pols:
     allocate(polar_dxyz(ubound(pols,1),ubound(pols,2),1))
     if(ctrl%maxsym) call create_red_pol(pols(:,:,:,1),dist_index, &
       & dsymm_index,nirred_atoms,&
       & irred_atom_index,2*ctrl%numderiv,char_table)
     call create_polar_dxyz(polar_dxyz(:,:,1),pols(:,:,:,1),distortion)
     call expand_polar(polar_dxyz(:,:,1),representation,symm_index,char_table)
     call symm_polar(polar_dxyz(:,:,1),representation,char_table)
     if(ctrl%print_pol) then
       call printmat(polar_dxyz(:,:,1),text='gradient of polarizability tensor',&
          & unit=out)
     end if
! G/LAO tensor:
     allocate(gtenlao_dxyz(ubound(gtenlao,1),ubound(gtenlao,2),1))
     if(ctrl%maxsym) call create_red_gten(gtenlao(:,:,:,1),dist_index, &
       & dsymm_index,nirred_atoms,&
       & irred_atom_index,2*ctrl%numderiv,char_table)
     call create_aten_dxyz(gtenlao_dxyz(:,:,1),gtenlao(:,:,:,1),distortion)
!MR C1 symmetry is expected, the following has never been tested:
!MR     call expand_gten(gtenlao_dxyz(:,:,1),representation,symm_index,char_table)
!MR     call symm_gten(gtenlao_dxyz(:,:,1),representation,char_table)
     if(ctrl%print_gaten) then
       call printmat(gtenlao_dxyz(:,:,1),text='gradient of G/LAO tensor',&
          & unit=out)
     end if
! G/AO tensor:
     allocate(gtenao_dxyz(ubound(gtenao,1),ubound(gtenao,2),1))
     if(ctrl%maxsym) call create_red_gten(gtenao(:,:,:,1),dist_index, &
       & dsymm_index,nirred_atoms,&
       & irred_atom_index,2*ctrl%numderiv,char_table)
     call create_aten_dxyz(gtenao_dxyz(:,:,1),gtenao(:,:,:,1),distortion)
!MR C1 symmetry is expected, the following has never been tested:
!MR     call expand_gten(gtenao_dxyz(:,:,1),representation,symm_index,char_table)
!MR     call symm_gten(gtenao_dxyz(:,:,1),representation,char_table)
     if(ctrl%print_gaten) then
       call printmat(gtenao_dxyz(:,:,1),text='gradient of G/AO tensor',&
          & unit=out)
     end if
! A tensor:
     allocate(aten_dxyz(ubound(aten,1),ubound(aten,2),1))
     call create_aten_dxyz(aten_dxyz(:,:,1),aten(:,:,:,1),distortion)
     if(ctrl%print_gaten) then
       call printmat(aten_dxyz(:,:,1),text='gradient of A tensor',&
          & unit=out)
     end if
  else if(ctrl%vroaew.ge.1.and.associated(ctrl%dynfreq)) then
! pols:
     allocate(polar_dxyz(ubound(pols,1),ubound(pols,2),ubound(ctrl%dynfreq,1)))
     do i=1,ubound(ctrl%dynfreq,1)
       if(ctrl%maxsym) call create_red_pol(pols(:,:,:,i),dist_index, &
          & dsymm_index,nirred_atoms,&
          & irred_atom_index,2*ctrl%numderiv,char_table)
       call create_polar_dxyz(polar_dxyz(:,:,i),pols(:,:,:,i),distortion)
       call expand_polar(polar_dxyz(:,:,i),representation,symm_index,char_table)
       call symm_polar(polar_dxyz(:,:,i),representation,char_table)
       if(ctrl%print_pol) then
         write(out,*)
         write(out,*)
         if(abs(ctrl%dynfreq(i)).le.1.0d-6) then
           write(out,'(a)') ' Dynamic polarizability derivatives for &
            &laser wavelength:  infinity'
         else
           write(out,'(a,f16.5)') ' Dynamic polarizability derivatives for &
            &laser wavelength: ',1.0d0/ctrl%dynfreq(i)*nm2freq
         end if
         write(out,*)
         call printmat(polar_dxyz(:,:,i),text='gradient of polarizability tensor',&
          & unit=out)
       end if
     end do
! G/LAO tensor:
     allocate(gtenlao_dxyz(ubound(gtenlao,1),ubound(gtenlao,2),ubound(ctrl%dynfreq,1)))
     do i=1,ubound(ctrl%dynfreq,1)
       if(ctrl%maxsym) call create_red_gten(gtenlao(:,:,:,i),dist_index, &
          & dsymm_index,nirred_atoms,&
          & irred_atom_index,2*ctrl%numderiv,char_table)
       call create_aten_dxyz(gtenlao_dxyz(:,:,i),gtenlao(:,:,:,i),distortion)
!MR C1 symmetry is expected, the following has never been tested:
!MR       call expand_gten(gtenlao_dxyz(:,:,i),representation,symm_index,char_table)
!MR       call symm_gten(gtenlao_dxyz(:,:,i),representation,char_table)
       if(ctrl%print_gaten) then
         write(out,*)
         write(out,*)
         if(abs(ctrl%dynfreq(i)).le.1.0d-6) then
           write(out,'(a)') ' Freq.-dependent G/LAO tensor derivatives for &
            &laser wavelength:  infinity'
         else
           write(out,'(a,f16.5)') ' Freq.-dependent G/LAO tensor derivatives for &
            &laser wavelength: ',1.0d0/ctrl%dynfreq(i)*nm2freq
         end if
         write(out,*)
         call printmat(gtenlao_dxyz(:,:,i),text='gradient of G/LAO tensor',&
          & unit=out)
       end if
     end do
! G/AO tensor:
     allocate(gtenao_dxyz(ubound(gtenao,1),ubound(gtenao,2),ubound(ctrl%dynfreq,1)))
     do i=1,ubound(ctrl%dynfreq,1)
       if(ctrl%maxsym) call create_red_gten(gtenao(:,:,:,i),dist_index, &
          & dsymm_index,nirred_atoms,&
          & irred_atom_index,2*ctrl%numderiv,char_table)
       call create_aten_dxyz(gtenao_dxyz(:,:,i),gtenao(:,:,:,i),distortion)
!MR C1 symmetry is expected, the following has never been tested:
!MR       call expand_gten(gtenao_dxyz(:,:,i),representation,symm_index,char_table)
!MR       call symm_gten(gtenao_dxyz(:,:,i),representation,char_table)
       if(ctrl%print_gaten) then
         write(out,*)
         write(out,*)
         if(abs(ctrl%dynfreq(i)).le.1.0d-6) then
           write(out,'(a)') ' Freq.-dependent G/AO tensor derivatives for &
            &laser wavelength:  infinity'
         else
           write(out,'(a,f16.5)') ' Freq.-dependent G/AO tensor derivatives for &
            &laser wavelength: ',1.0d0/ctrl%dynfreq(i)*nm2freq
         end if
         write(out,*)
         call printmat(gtenao_dxyz(:,:,i),text='gradient of G/AO tensor',&
          & unit=out)
       end if
     end do
! A tensor:
     allocate(aten_dxyz(ubound(aten,1),ubound(aten,2),ubound(ctrl%dynfreq,1)))
     do i=1,ubound(ctrl%dynfreq,1)
       call create_aten_dxyz(aten_dxyz(:,:,i),aten(:,:,:,i),distortion)
       if(ctrl%print_gaten) then
         write(out,*)
         write(out,*)
         if(abs(ctrl%dynfreq(i)).le.1.0d-6) then
           write(out,'(a)') ' Freq.-dependent A tensor derivatives for &
            &laser wavelength:  infinity'
         else
           write(out,'(a,f16.5)') ' Freq.-dependent A tensor derivatives for &
            &laser wavelength: ',1.0d0/ctrl%dynfreq(i)*nm2freq
         end if
         write(out,*)
         call printmat(aten_dxyz(:,:,i),text='gradient of A tensor',&
          & unit=out)
       end if
     end do
  end if

! allocation of arrays for VROA calculations finished
!#####################################################################################
!
  if(ctrl%print_cioslowski) then
     call create_cioslowski(dipol_dxyz,cioslowski)
     call print_cioslowski(out,cioslowski)
     deallocate(cioslowski)
  endif
  !----------------------------------------------------------------
  ! continue in ctrl%isosym symmetry if unusual isotope settings appear
  !----------------------------------------------------------------
  !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
100 if(any(molecule(:)%changem).or.ctrl%isosym.ne.'def') then
     if(ctrl%isosym.eq.'def') then
        call get_point_grp(molecule,1.D-3,isotope_equal,ctrl%isosym)
     endif
     write(out,*)
     write(out,'(''======================================================='')')
     write(out,'(''======================================================='')')
     write(out,'('' ALL FOLLOWING RESULTS WERE OBTAINED USING '',a3,'' SYMMETRY'')')  ctrl%isosym
     write(out,'(''======================================================='')')
     write(out,'(''======================================================='')')
     write(out,*)
     deallocate(proj_op,representation,character,irreps)
     call destroy_char_table(char_table)
     call create_char_table(char_table,ctrl%isosym)
     call create_representation_for_vib(representation,molecule,char_table)
     call create_character(representation,character)
     call create_irreps(character,irreps,char_table)
     call create_proj_op(representation,char_table,proj_op)
  endif
  !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  !----------------------------------------------------------------

  !----------------------------------------------------------------
  ! test linearity of molecule
  !----------------------------------------------------------------
  if(testlin(molecule)) call serror('snf',-1,'Found linear molecule.',out)
  !----------------------------------------------------------------
  ! Header printed here, so that it is printed only once.
  !----------------------------------------------------------------
  write(out,*)
  write(out,'('' Identifying translational and rotation normal modes'')')
  write(out,'('' ---------------------------------------------------'')')
!--------------------------------------------------------------------
! intensities-only-mode: set symmetry to c1
! (may be changed in future versions to automatic
! symmetry check)
!--------------------------------------------------------------------
1000  if(ctrl_restart%nummodes.ne.0) then
        char_table%nirrep=1
        allocate(char_table%mulliken(1),char_table%g(1))
        char_table%mulliken(1) = '  a  '
        char_table%g(1)=1
        irrep_dim=ctrl_restart%nummodes
          allocate(  &
  &   spectrum(ctrl_restart%nummodes), & 
  &    norm_modes(nvar,ctrl_restart%nummodes))
       allocate(global_spectrum(ctrl_restart%nummodes))
       global_spectrum(:)%g=1

  !----------------------------------------------------------------
  ! solve problem for every irreducible representation separately
  !----------------------------------------------------------------
  else
    allocate(global_spectrum(nvar-number_projected_modes))
  end if
  if(ctrl%ramanew.ge.1.or.ctrl%vroaew.ge.1) then
    do i=1,ubound(global_spectrum,1)
      allocate(global_spectrum(i)%raman(nfreq))
    end do
  end if
  offset=0
!--------------------------------------------------------------------
  do irrep=1,char_table%nirrep
!--------------------------------------------------------------------
! intensities-only-mode: skip diagonalization of Hessian
!--------------------------------------------------------------------
     if(ctrl_restart%nummodes.ne.0) then 
       goto 1100
     end if

     irrep_dim=irreps(irrep)*char_table%l(1,irrep)
     if(irrep_dim .ne. 0) then
        lwork=(ilaenv(1,'DSYTRD','l',nvar,nvar,nvar,nvar)&
             & +2)*nvar
        allocate(eig_proj_op(nvar),work(lwork))
        !----------------------------------------------------------------
        ! calculate eigenvalues and eigenvectors of the 
        ! projection operators.
        !----------------------------------------------------------------
        call dsyev('V','L',nvar,proj_op(1,1,irrep),nvar,eig_proj_op(1),work(1),lwork,info)
        if(info.ne.0) then 
          call serror('snf',2,'Could not calculate eigenvectors##&
           &of the projection operators.',out)
        end if
        deallocate(eig_proj_op,work)

        allocate(c2p(nvar,irrep_dim))
        c2p=proj_op(1:nvar,nvar-irrep_dim+1:nvar,irrep)
        !----------------------------------------------------------------
        ! do everything twice to get an assessment of the quality of 
        ! rotational and translational modes
        ! you may comment it out, if you want. nothing relies on this.
        !----------------------------------------------------------------
        lwork=(ilaenv(1,'DSYTRD','l',irrep_dim,irrep_dim,irrep_dim,irrep_dim)&
             & +2)*irrep_dim
        allocate(hessian_p(irrep_dim,irrep_dim),&
             &   hessian_q(irrep_dim),&
             &   spectrum(irrep_dim), &
             &   work(lwork), &
             &   tmpmat(nvar,irrep_dim), &
             &   norm_modes(nvar,irrep_dim))
        tmpmat=0.D0
        !----------------------------------------------------------------
        ! "block diagonalization" 
        ! [S_mu^(m)]^+ F^(m) S_mu^(m) = F_mu^(s)
        !----------------------------------------------------------------
        call dgemm('N','N',nvar,irrep_dim,nvar,1.D0,hessian_m(1,1),nvar,c2p(1,1),nvar,0.D0,tmpmat(1,1),nvar)
        hessian_p=0.D0
        call dgemm('T','N',irrep_dim,irrep_dim,nvar,1.D0,c2p(1,1),nvar,tmpmat(1,1),nvar,0.D0,hessian_p(1,1),irrep_dim)
        call balance_hessian(hessian_p)
        !--------------------------------------------------------------------
        ! calculate eigenvalues and eigenvectors of this block of the hessian
        ! [U_mu^(s)]^+ F_mu^(s) U_mu^(s) = F_mu^(q)
        !--------------------------------------------------------------------
        call dsyev('V','L',irrep_dim,hessian_p(1,1),irrep_dim,hessian_q(1),work(1),lwork,info)
        if(info.ne.0) then 
          call serror('snf',2,'Could not calculate eigenvectors##&
           &of the hessian.',out)
        end if
        !--------------------------------------------------------------------
        ! calculate "original" transformation matrix
        ! U_mu^(m) = S_mu^(m) U_mu^(s) 
        !--------------------------------------------------------------------
        call dgemm('N','N',nvar,irrep_dim,irrep_dim,1.D0,c2p(1,1),nvar, &
         &hessian_p(1,1),irrep_dim,0.D0,norm_modes(1,1),nvar)
        call calc_frequencies(hessian_q,spectrum,ctrl%freqscale)
        do i=1,irrep_dim
           allocate(spectrum(i)%norm_mode(nvar))
           spectrum(i)%norm_mode(:)=norm_modes(:,i)
           spectrum(i)%irrep=char_table%mulliken(irrep)
           spectrum(i)%g=char_table%g(irrep)
           spectrum(i)%number_in_irrep=i
        end do
        call get_translation_and_rotation(spectrum,molecule,-1,unit=out)

        deallocate(norm_modes)
        call destroy_spectrum(spectrum)
        deallocate(hessian_p,hessian_q,work,tmpmat)
        !----------------------------------------------------------------
        ! end of rotational and translational analysis
        !----------------------------------------------------------------
        ! now project out all translational and rotational modes and 
        ! do the whole work again
        !----------------------------------------------------------------
        write(out,*)
        write(out,'(a)') ' Translational and rotational modes are now projected off!'
        write(out,'(a)') ' Frequencies may slightly change but the resulting data are more accurate!'
        call project_off_transrot(molecule,c2p,'all')

        irrep_dim=ubound(c2p,2)
     endif
     if(irrep_dim .ne. 0) then
        lwork=(ilaenv(1,'DSYTRD','l',nvar,nvar,nvar,nvar)&
             & +2)*nvar
        allocate(hessian_p(irrep_dim,irrep_dim),&
             &   hessian_q(irrep_dim),&
             &   spectrum(irrep_dim), &
             &   work(lwork), &
             &   tmpmat(nvar,irrep_dim), &
             &   norm_modes(nvar,irrep_dim))
        !----------------------------------------------------------------
        ! transform hessian to symmetry adapted basis
        ! and diagonalize it.
        !----------------------------------------------------------------
        call dgemm('N','N',nvar,irrep_dim,nvar,1.D0,hessian_m(1,1),nvar,c2p(1,1),nvar,0.D0,tmpmat(1,1),nvar)
        call dgemm('T','N',irrep_dim,irrep_dim,nvar,1.D0,c2p(1,1),nvar,tmpmat(1,1),nvar,0.D0,hessian_p(1,1),irrep_dim)
        call balance_hessian(hessian_p)
        !----------------------------------------------------------------
        ! calculate eigenvalues and eigenvectors of the symmetry
        ! adapted hessian
        !----------------------------------------------------------------
        call dsyev('V','L',irrep_dim,hessian_p(1,1),irrep_dim,hessian_q(1),work(1),lwork,info)
        if(info.ne.0) then 
          call serror('snf',2,'Could not calculate eigenvectors##&
           &of the symmetry adapted hessian.',out)
        end if
        !----------------------------------------------------------------
        ! create transformation matrix 'normal_modes' from 
        ! cartesian to normal coordinates
        !----------------------------------------------------------------
        call dgemm('N','N',nvar,irrep_dim,irrep_dim,1.D0,c2p(1,1),nvar,hessian_p(1,1),irrep_dim,0.D0,norm_modes(1,1),nvar)
        call calc_frequencies(hessian_q,spectrum,ctrl%freqscale)
!--------------------------------------------------------------------
! intensities-only mode: ... continue here after having skipped
! diagonalization of Hessian
!--------------------------------------------------------------------
1100 continue
        !----------------------------------------------------------------
        ! perform transformation of cartesian derivative of dipol moments
        ! calculate ir (and raman) spectra with eigenvalues and 
        ! transformed dipol derivatives
        !----------------------------------------------------------------
        ! don't do this in case of forward or backward differences
        !----------------------------------------------------------------
        if(ctrl%dtype.ne.'centr') cycle
                if(ctrl_restart%nummodes.eq.0) then
          call create_dipol_dq(dipol_dxyz,norm_modes,molecule,dipol_dq)
        else
          natoms=ubound(molecule,1)
          allocate(dipol_dq(3,ctrl%nummodes))
          ! mass-weigthing of normal modes: ctrl%nmodein --> normalmodes
          allocate(normmodes(3*natoms,3*natoms))
          normmodes=ctrl%nmodein
          call nmod_c2m(normmodes,natoms,molecule,ctrl%nummodes)
          call create_dipol_dq_intonly(dipol_dq,dipols,ctrl%nummodes, &
          &  normmodes,ctrl%cstep,ctrl%numderiv,natoms,molecule)
          spectrum(:)%freq = ctrl%freqin(:)
          if(associated(ctrl%freqin))  deallocate(ctrl%freqin)
          if(associated(ctrl%selectmode)) deallocate(ctrl%selectmode)

        end if
        call calc_ir_intensities(dipol_dq,spectrum)
        if ((ctrl%ramanew.ge.1.or.ctrl%vroaew.ge.1).and.(.not.associated(ctrl%dynfreq))) then
          allocate(polar_dq(6,ubound(norm_modes,2),1))
          do i=1,irrep_dim
            allocate(spectrum(i)%raman(nfreq))
          end do
          if(ctrl_restart%nummodes.eq.0) then
            call create_polar_dq(polar_dxyz(:,:,1),norm_modes,molecule,polar_dq(:,:,1))
          else
            call create_polar_dq_intonly(polar_dq(:,:,1),pols(:,:,:,1), & 
            & ctrl%nummodes,normmodes,ctrl%cstep,ctrl%numderiv,natoms,molecule)
          end if
          call calc_raman_intensities(polar_dq(:,:,1),spectrum,ctrl%temp,&
           &ctrl%theta,1) 
        else if((ctrl%ramanew.ge.1.or.ctrl%vroaew.ge.1).and.associated(ctrl%dynfreq)) then
          allocate(polar_dq(6,ubound(norm_modes,2),nfreq))
          do i=1,irrep_dim
            allocate(spectrum(i)%raman(nfreq))
          end do
          do i=1,ubound(ctrl%dynfreq,1)
            if(ctrl_restart%nummodes.eq.0) then
              call create_polar_dq(polar_dxyz(:,:,i),norm_modes,molecule,polar_dq(:,:,i))
            else
              call create_polar_dq_intonly(polar_dq(:,:,1),pols(:,:,:,1), &
              & ctrl%nummodes,normmodes,ctrl%cstep,ctrl%numderiv,natoms,molecule)
            end if
            call calc_raman_intensities(polar_dq(:,:,i),spectrum,ctrl%temp,&
             &ctrl%theta,i,ctrl%dynfreq(i)) 
          end do
        else 
          !----------------------------------------------------------------
          ! This is done for compatibility with other parts of the program
          !----------------------------------------------------------------
          do i=1,irrep_dim
            allocate(spectrum(i)%raman(1))
          end do
        endif
        !----------------------------------------------------------------
        ! Generate derivatives of G and A tensors for VROA w.r.t. normal coordinates
        !----------------------------------------------------------------
        if (ctrl%vroaew.ge.1.and.(.not.associated(ctrl%dynfreq))) then
          allocate(polar9_dq(9,ubound(norm_modes,2),1))
!... copy 6 non-redundant entries of polarizability derivatives to 9 redundant entries
!    and transfrom back to atomic units (mass AND length!)
          ddum=sqrt(XFAMU)*bohr2ang*bohr2ang
          polar9_dq(1,:,1)=polar_dq(1,:,1)/ddum
          polar9_dq(2,:,1)=polar_dq(2,:,1)/ddum
          polar9_dq(3,:,1)=polar_dq(3,:,1)/ddum
          polar9_dq(4,:,1)=polar_dq(2,:,1)/ddum
          polar9_dq(5,:,1)=polar_dq(4,:,1)/ddum
          polar9_dq(6,:,1)=polar_dq(5,:,1)/ddum
          polar9_dq(7,:,1)=polar_dq(3,:,1)/ddum
          polar9_dq(8,:,1)=polar_dq(5,:,1)/ddum
          polar9_dq(9,:,1)=polar_dq(6,:,1)/ddum
          allocate(gtenlao_dq(9,ubound(norm_modes,2),1))
          allocate(gtenao_dq(9,ubound(norm_modes,2),1))
          allocate(aten_dq(27,ubound(norm_modes,2),1))
!MR type spectrum does not yet contain vroa as a possible mode
!MR          do i=1,irrep_dim
!MR            allocate(spectrum(i)%vroa(nfreq))
!MR          end do
          if(ctrl_restart%nummodes.eq.0) then
            call create_aten_dq(gtenlao_dxyz(:,:,1),norm_modes,molecule,gtenlao_dq(:,:,1))
            call create_aten_dq(gtenao_dxyz(:,:,1),norm_modes,molecule,gtenao_dq(:,:,1))
            call create_aten_dq(aten_dxyz(:,:,1),norm_modes,molecule,aten_dq(:,:,1))
          else
            call create_aten_dq_intonly(gtenlao_dq(:,:,1),gtenlao(:,:,:,1),ctrl%nummodes, & 
            & normmodes,ctrl%cstep,ctrl%numderiv,natoms,molecule)
            call create_aten_dq_intonly(gtenao_dq(:,:,1),gtenao(:,:,:,1),ctrl%nummodes,&
            & normmodes,ctrl%cstep,ctrl%numderiv,natoms,molecule)
            call create_aten_dq_intonly(aten_dq(:,:,1),aten(:,:,:,1),ctrl%nummodes, &
            & normmodes,ctrl%cstep,ctrl%numderiv,natoms,molecule)
          end if
!... transfrom mass dimension to atomic units
          ddum=sqrt(XFAMU)
          gtenlao_dq(:,:,1)=gtenlao_dq(:,:,1)/ddum
          gtenao_dq(:,:,1)=gtenao_dq(:,:,1)/ddum
          aten_dq(:,:,1)=aten_dq(:,:,1)/ddum
          write(out,'(a)') ' '
          write(out,'(a)') ' Entering routines based on the DALTON program system (ABACUS) for VROA calcs:'
          write(out,'(a)') ' ... original version of the routines by K. Ruud et al. !'
          write(out,'(a)') ' ... see: http://www.kjemi.uio.no/software/dalton/dalton.html'
          write(out,'(a)') ' '
          allocate(freqs(ubound(spectrum,1)))
          do i=1,ubound(spectrum,1)
            freqs(i)=spectrum(i)%freq
          enddo
          dumlaserf(1)=0.0d0
!SL
          call calcvroa(freqs,polar9_dq,gtenao_dq,gtenlao_dq,aten_dq,&
            &ubound(norm_modes,1),ubound(norm_modes,1),1,1,dumlaserf,out,&
              & ctrl%Atensor_0,ctrl%progno)
          deallocate(freqs)
        else if(ctrl%vroaew.ge.1.and.associated(ctrl%dynfreq)) then
          allocate(polar9_dq(9,ubound(norm_modes,2),nfreq))
!... copy 6 non-redundant entries of polarizability derivatives to 9 redundant entries
!    and transfrom back to atomic units (mass AND length!)
          ddum=sqrt(XFAMU)*bohr2ang*bohr2ang
          polar9_dq(1,:,:)=polar_dq(1,:,:)/ddum
          polar9_dq(2,:,:)=polar_dq(2,:,:)/ddum
          polar9_dq(3,:,:)=polar_dq(3,:,:)/ddum
          polar9_dq(4,:,:)=polar_dq(2,:,:)/ddum
          polar9_dq(5,:,:)=polar_dq(4,:,:)/ddum
          polar9_dq(6,:,:)=polar_dq(5,:,:)/ddum
          polar9_dq(7,:,:)=polar_dq(3,:,:)/ddum
          polar9_dq(8,:,:)=polar_dq(5,:,:)/ddum
          polar9_dq(9,:,:)=polar_dq(6,:,:)/ddum
          allocate(gtenlao_dq(9,ubound(norm_modes,2),nfreq))
          allocate(gtenao_dq(9,ubound(norm_modes,2),nfreq))
          allocate(aten_dq(27,ubound(norm_modes,2),nfreq))
!MR type spectrum does not yet contain vroa as a possible mode
!MR          do i=1,irrep_dim
!MR            allocate(spectrum(i)%vroa(nfreq))
!MR          end do
          do i=1,ubound(ctrl%dynfreq,1)
          if(ctrl_restart%nummodes.eq.0) then
            call create_aten_dq(gtenlao_dxyz(:,:,1),norm_modes,molecule,gtenlao_dq(:,:,1))
            call create_aten_dq(gtenao_dxyz(:,:,1),norm_modes,molecule,gtenao_dq(:,:,1))
            call create_aten_dq(aten_dxyz(:,:,1),norm_modes,molecule,aten_dq(:,:,1))
          else
            call create_aten_dq_intonly(gtenlao_dq(:,:,1),gtenlao(:,:,:,1),ctrl%nummodes, & 
            & normmodes,ctrl%cstep,ctrl%numderiv,natoms,molecule)
            call create_aten_dq_intonly(gtenao_dq(:,:,1),gtenao(:,:,:,1),ctrl%nummodes,&
            & normmodes,ctrl%cstep,ctrl%numderiv,natoms,molecule)
            call create_aten_dq_intonly(aten_dq(:,:,1),aten(:,:,:,1),ctrl%nummodes, &
            & normmodes,ctrl%cstep,ctrl%numderiv,natoms,molecule)
          end if
!... transfrom mass dimension to atomic units
            ddum=sqrt(XFAMU)
            gtenlao_dq(:,:,i)=gtenlao_dq(:,:,i)/ddum
            gtenao_dq(:,:,i)=gtenao_dq(:,:,i)/ddum
            aten_dq(:,:,i)=aten_dq(:,:,i)/ddum
          end do
          write(out,'(a)') ' '
          write(out,'(a)') ' Entering routines based on the DALTON program system (ABACUS) for VROA calcs:'
          write(out,'(a)') ' ... original version of the routines by K. Ruud et al. !'
          write(out,'(a)') ' ... see: http://www.kjemi.uio.no/software/dalton/dalton.html'
          write(out,'(a)') ' '
          allocate(freqs(ubound(spectrum,1)))
	  do i=1,ubound(spectrum,1)
            freqs(i)=spectrum(i)%freq
          enddo
!SL
          call calcvroa(freqs,polar9_dq,gtenao_dq,gtenlao_dq,aten_dq,&
                       &ubound(norm_modes,2),ubound(norm_modes,2),nfreq,&
                       &nfreq,ctrl%dynfreq,out,ctrl%Atensor_0,ctrl%progno)
          deallocate(freqs)
!MR        else
!MR          do i=1,irrep_dim
!MR            allocate(spectrum(i)%vroa(1))
!MR          end do
        endif
! intensities-only mode: some reassignment of normal modes part I...
  if(ctrl_restart%nummodes.ne.0) then
    ctrl%nmodein=normmodes
    if(allocated(normmodes)) deallocate(normmodes)
  end if
        !----------------------------------------------------------------
        ! assign several values concerning the spectrum
        !----------------------------------------------------------------
        do i=1,irrep_dim
           allocate(spectrum(i)%norm_mode(nvar))
           if(ctrl_restart%nummodes.ne.0) then
! intensities-only mode: some reassignment of normal modes part II...
             spectrum(i)%norm_mode(:)=ctrl%nmodein(i,:)
           else
             spectrum(i)%norm_mode(:)=norm_modes(:,i)
           end if
           spectrum(i)%irrep=char_table%mulliken(irrep)
           spectrum(i)%g=char_table%g(irrep)
           spectrum(i)%number_in_irrep=i
           spectrum(i)%red_mass = 0.d0
           do j=1,nvar
              spectrum(i)%red_mass = spectrum(i)%red_mass + &
                   & (spectrum(i)%norm_mode(j)**2 &
                   & /(molecule((j+2)/3)%mass))
           enddo
           spectrum(i)%red_mass=1.d0/spectrum(i)%red_mass
           spectrum(i)%force = 0.d0
        enddo
        if (allocated(hessian_q)) deallocate(hessian_q)
        if (allocated(hessian_p)) deallocate(hessian_p)
        if (associated(dipol_dq)) deallocate(dipol_dq)
        if (allocated(work)) deallocate(work)
        if (allocated(tmpmat)) deallocate(tmpmat)
        if (allocated(norm_modes)) deallocate(norm_modes)
        if (associated(polar_dq)) deallocate(polar_dq)
! set free all memory used for VROA tensors:
        if (associated(polar9_dq)) deallocate(polar9_dq)
        if (associated(gtenlao_dq)) deallocate(gtenlao_dq)
        if (associated(gtenao_dq)) deallocate(gtenao_dq)
        if (associated(aten_dq)) deallocate(aten_dq)
        if (associated(gtenlao_dxyz)) deallocate(gtenlao_dxyz)
        if (associated(gtenao_dxyz)) deallocate(gtenao_dxyz)
        if (associated(aten_dxyz)) deallocate(aten_dxyz)

        !----------------------------------------------------------------
        ! copy irrep data to global spectrum
        !----------------------------------------------------------------
        global_spectrum(offset+1:offset+irrep_dim)=spectrum(:)
        offset=offset+irrep_dim
        deallocate(spectrum)
     endif
     if(associated(c2p)) deallocate(c2p)
  enddo ! loop over irreps
  ! In case of unexpected incorrect translational or rotational modes
  ! dummy modes are introduced
! intensities-only mode: skip some irelevant assignments
  if(ctrl_restart%nummodes.ne.0) goto 1150
  do i=offset+1,nvar-number_projected_modes
     allocate(global_spectrum(i)%norm_mode(nvar))
     global_spectrum(i)%norm_mode = 0.D0
     global_spectrum(i)%freq = 0.D0
     global_spectrum(i)%irrep='ERR'
     global_spectrum(i)%g=0
     global_spectrum(i)%number_in_irrep = 0
     if(ctrl%ramanew.ge.1.or.ctrl%vroaew.ge.1) then 
       allocate(global_spectrum(i)%raman(nfreq))
       global_spectrum(i)%raman(:)%int=0.D0
     end if
     global_spectrum(i)%red_mass=0.D0
  enddo
  if(ctrl%dtype.ne.'centr') stop 'No intensities are computed in this testrun'
  !----------------------------------------------------------------
  ! clean up, and print out.
  !----------------------------------------------------------------
  1150 continue
  call sort_spectrum(global_spectrum)
  call assign_irrep_number(global_spectrum,char_table)
  call get_relir(global_spectrum)
  call print_ir(global_spectrum,unit=out)
  static=.not.associated(ctrl%dynfreq)
  if((ctrl%ramanew.ge.1.or.ctrl%vroaew.ge.1).and.(.not.associated(ctrl%dynfreq))) then
    call get_relcross(global_spectrum,1)
    call print_raman(global_spectrum,1,out)
    if(ctrl%print_rcs) call print_rcs(global_spectrum,1,static,&
    &ctrl%theta,out)
  else if((ctrl%ramanew.ge.1.or.ctrl%vroaew.ge.1).and.associated(ctrl%dynfreq)) then
    !----------------------------------------------------------------
    ! if dynamic polarizabilities with non-standard laser wavelengths
    ! have been selected, print raman intensities for these
    ! wavelengths.
    !----------------------------------------------------------------
    do i=1,nfreq
      if(abs(ctrl%dynfreq(i)).le.1.0d-6) then
        write(out,'(3(/),a)') ' Following Raman results obtained for &
         &laser wavelength:  infinity'
        write(out,'(80a)') ('=',j=1,80)
        static=.true.
      else
        write(out,'(3(/),a,f16.5,a)') ' Following Raman results obtained &
         &for laser wavelength: ',1.0d0/ctrl%dynfreq(i)*nm2freq,' nm'
         write(out,'(80a)') ('=',j=1,80)
        static=.false.
      end if
      write(out,'(a)') 
      call get_relcross(global_spectrum,i)
      call print_raman(global_spectrum,i,out)
      if(ctrl%print_rcs) call print_rcs(global_spectrum,i,static,&
       &ctrl%theta,out)
    end do
  end if
  !-------------------------------------------------------------
  ! additional output: intensities summed over degenerate modes
  !-------------------------------------------------------------
  if(ctrl%sumdeg) then
    write(out,'((/),152(''#''))') 
    write(out,'(3(/),a)') ' Intensities summed over degenerate modes'
    write(out,'(a,2(/))') ' ----------------------------------------'
    call sum_degen(global_spectrum,sum_spectrum,rflag)
    call get_relir(sum_spectrum)
    call print_ir(sum_spectrum,unit=out)
    if((ctrl%ramanew.ge.1.or.ctrl%vroaew.ge.1).and.(.not.associated(ctrl%dynfreq))) then
      call get_relcross(sum_spectrum,1)
      call print_raman(sum_spectrum,1,out)
      if(ctrl%print_rcs) call print_rcs(sum_spectrum,1,static,&
       &ctrl%theta,out)
!CMR291004    else if((ctrl%ramanew.ge.1.or.ctrl%vroaew.ge.1).and.associated(ctrl%dynfreq)) then
!CMR291004 It was necessary to deactivate this for VROA because parts of sum_spectrum
!CMR291004   have not been allocated up to this point (namely: sum_spectrum%raman(*),
!CMR291004   which will cause a segmentation fault in call get_relcross(sum_spectrum,i)
!CMR291004
    else if(ctrl%vroaew.ge.1.and.associated(ctrl%dynfreq)) then
          call serror('snf',1,'No intensity summation over degenerate modes&
     & ##available for VROA at the moment!.',out) 
!CMR291004
    else if(ctrl%ramanew.ge.1.and.associated(ctrl%dynfreq)) then
      do i=1,nfreq
        if(abs(ctrl%dynfreq(i)).le. 1.0d-10) then
          write(out,'(3(/),a)') ' Following Raman results obtained for &
           &laser wavelength:  infinity'
        else
          write(out,'(3(/),a,f16.5)') ' Following Raman results obtained &
           &for laser wavelength: ',1.0d0/ctrl%dynfreq(i)*nm2freq
        end if
        write(out,'(a)') 
        call get_relcross(sum_spectrum,i)
        call print_raman(sum_spectrum,i,out)
        if(ctrl%print_rcs) call print_rcs(sum_spectrum,i,static,&
         &ctrl%theta,out)
      end do
    end if
    deallocate(sum_spectrum)
    write(out,'(2(/),152(''#''),2(/))') 
  end if
  !-------------------------------------
  ! thermochemistry
  !-------------------------------------
! intensities-only modes: skip thermochemistry
  if(ctrl_restart%nummodes.ne.0) goto 1200
  molekel%mass=0.0d0
  do k=1,ubound(atom_index,1)
     molekel%mass=molekel%mass+molecule(k)%mass
  enddo
  mol_mass=molekel%mass/(1000.0d0*avogadro)
  if(ctrl%print_zpe) then
     call zeropoint_energy(global_spectrum,zeropoint)
     write(out,*)
     write(out,'('' General molecular quantities'')')
     write(out,'('' ----------------------------'')')
     write(out,'(''   Zero point energy (ZPE) [kJ/mol]   :'',F10.1)') zeropoint
     write(out,'(''   Molecular mass [amu]               :'',F10.1)') molekel%mass
     write(out,'(''   Principal momenta and eigenvectors :'')')
     write(out,*)
     write(out,'(3F15.5)')(molekel%p_inertia(i),i=1,3)
     write(out,*)
     do j=1,3
        write(out,'(3F15.5)')(molekel%p_axes(i,j),i=1,3)
     enddo
  endif
  !-------------------------------------------------------------------
  ! checking symmetry number (some are not sufficiently tested)
  !-------------------------------------------------------------------
  select case(ctrl%symmetry)
    case('th ','o  ','i  ','ih ')
    call serror('snf',-1,'Note that the symmetry numbers have&
     & not been##sufficiently checked for this point group.',out)
  end select
  if(testlin(molecule)) then
    if(index(char_table%point_group(1:3),'h').ne.0) then
      char_table%sigma=2 
    else 
      char_table%sigma=1 
    end if
  end if
  if(ctrl%print_thermo) then
     allocate(freqs(ubound(global_spectrum,1)))
     do i=1,ubound(global_spectrum,1)
        freqs(i)=global_spectrum(i)%freq
     enddo
     call get_tdvalues(zeropoint,ctrl%elec_energy,freqs,ctrl%temp,ctrl%press,out,&
      & mol_mass,molekel%p_inertia,char_table%sigma,testlin(molecule),ctrl%spin)
     deallocate(freqs)
  endif
  !cmr
! intensities-only modes: skip normal-mode output
  if(ctrl_restart%nummodes.ne.0) goto 1200
  write(out,*)
  write(out,'('' Frequencies and mass-weighted normal coordinates'')')
  write(out,'('' ------------------------------------------------'')')
  call preig(out,ubound(global_spectrum,1),global_spectrum,'eigenvalues  in  cm^-1')
  call print_norm_modes(global_spectrum,molecule,unit=out)
  !----------------------------------------------------------------
  ! warning message in case of wrong hessian
  !----------------------------------------------------------------
  if(new_restart) then
     write(out,*)'*********************************************************'
     write(out,*)'                     W A R N I N G'
     write(out,*)'*********************************************************'
     write(out,*)' Calculation is not valid due to an unsymmetric hessian.'
     write(out,*)' Decide whether the data obtained here are sufficiently'
     write(out,*)' accurate for further use. In case of doubt, restart the'
     write(out,*)' calculation with the new restart file. The original one'
     write(out,*)' is stored on '//trim(adjustl(ctrl%restart))//'.bak.'
     write(out,*)'*********************************************************'
  endif
  !----------------------------------------------------------------
  ! write mopac and gaussian frequency output
  !----------------------------------------------------------------
  call calc_mode_properties(molecule,global_spectrum)
1200 continue
  if(ctrl%print_mopac.or.ctrl%print_gaussian.or.ctrl%print_xmol) then
     write(out,*)
     write(out,'('' Generate fake outputs for normal mode presentation'')')
     write(out,'('' --------------------------------------------------'')')
     if((ctrl%ramanew.ge.1.or.ctrl%vroaew.ge.1).and.associated(ctrl%dynfreq)) then
       if(ubound(ctrl%dynfreq,1).ge.2) call serror('snf',-1,'Files for&
         & normal mode presentation contain##Raman data only for the first&
         & selected frequency.',out)
     end if
     if(ctrl%print_mopac) &
          &  call print_mopac_frequencies(molecule,global_spectrum,&
          &            ctrl%print_mopxmol,'mopac.out',unit=out)
     if(ctrl%print_gaussian) &
          &  call print_gaussian_frequencies(molecule,global_spectrum,&
          &                            'g98.out',rflag,unit=out)       
     if(ctrl%print_xmol) &
      & call print_xmol_norm(molecule,global_spectrum,'xmol.XYZ.out',&
      &ctrl%xmolsc1,unit=out)
  endif
  !----------------------------------------------------------------
  ! generate data-, gnuplot-, and eps-files for spectrum plotting
  !----------------------------------------------------------------
  call generate_plots(ctrl,global_spectrum,1,unit=out)
  !----------------------------------------------------------------
  ! clean up memory 
  !----------------------------------------------------------------
  call destroy_spectrum(global_spectrum)
  deallocate(global_spectrum)
  call destroy_char_table(char_table)
  if(associated(representation)) deallocate(representation)
  if(associated(character)) deallocate(character)
  if(associated(irreps)) deallocate(irreps)
  if(associated(proj_op)) deallocate(proj_op)
  if(associated(dipol_dxyz)) deallocate(dipol_dxyz)
  if(associated(hessian_xyz)) deallocate(hessian_xyz)
  if(allocated(hessian_m)) deallocate(hessian_m)
  call destroy_molecule(molecule)
  call destroy_irred_atoms(atom_index,symm_index,irred_atom_index)
  call destroy_control(ctrl)
  if(allocated(distortion)) deallocate(distortion)
!=============
! time control
!=============
  call secs(t2)
  call clean(tbuf)
  call ptime(t2-t1,tbuf)
  call llen(tbuf,ltbuf)
  write(out,*)  
  write(out,'(500a)')  ('*',j=1,ltbuf+3+21)
  write(out,'(3a)')'** time used in snf: ',tbuf(1:ltbuf),' **'
  write(out,'(500a)')  ('*',j=1,ltbuf+3+21)
!=============
  close(out)
end program snf

subroutine print_header(out)
  implicit none
  include '../snfdc/VERSION.h'
  integer,intent(in) :: out
  write(out,'(A)')
!                   +1234567890123456789012345678+8765432109876543210987654321+'
  write(out,'(A)')' +---------------------------------------------------------+'
  write(out,'(A)')' |                                                         |'
  write(out,'(A)')' |                          S N F                          |'
  write(out,'(A)')' |                                                         |'
  write(out,'(A,a21,a)') &
   &              ' |                  ',version,          '                  |'
  write(out,'(A)')' |                                                         |'
  write(out,'(A)')' |   A program package for numerical frequency analyses    |'
  write(out,'(A)')' |                                                         |'
  write(out,'(A)')' |                       written by                        |'
  write(out,'(A)')' |                                                         |'
  write(out,'(A)')' |               C. Herrmann, J. Neugebauer,               |'
  write(out,'(A)')' |      C. Kind, S. Luber, B. A. Hess, and M. Reiher       |'
  write(out,'(A)')' |                                                         |'
  write(out,'(A)')' |           University of Erlangen, 1999 - 2003           |'
  write(out,'(A)')' |           University of Bonn,     2004 - 2005           |'
  write(out,'(A)')' |           University of Jena,     2005 - 2006           |'
  write(out,'(A)')' |                                                         |'
  write(out,'(A)')' |   Laboratorium fuer Physikalische Chemie, ETH Zurich,   |'
  write(out,'(A)')' |    Wolfgang-Pauli-Str. 10, 8093 Zurich, Switzerland     |'
  write(out,'(A)')' |                       2006 - 2007                       |'
  write(out,'(A)')' |                                                         |'
  write(out,'(A)')' +---------------------------------------------------------+'
!                   +1234567890123456789012345678+8765432109876543210987654321+'
  write(out,'(A)')
end subroutine print_header

subroutine last_mod(out)
  implicit none
  include '../snfdc/VERSION.h'
  integer,intent(in) :: out
  write(out,'(A)')
  write(out,'(2A)') ' Last modification: ',lastmod
  write(out,'(A)')
end subroutine last_mod
