
!********************************************************************************
!*   Copyright 1999 -- 2002 C. Kind, M. Reiher, J. Neugebauer, and B. A. Hess,
!*   Chair of Theoretical Chemistry, Friedrich-Alexander-University 
!*   Erlangen-Nuremberg, Egerlandstr. 3, D-91058 Erlangen, Germany.
!*
!*
!*   This file is part of the package SNF.
!*
!*   SNF is free software; you can redistribute it and/or modify
!*   it under the terms of the GNU General Public License as published by
!*   the Free Software Foundation; either version 2 of the License, or
!*   (at your option) any later version.
!*
!*   This program is distributed in the hope that it will be useful,
!*   but WITHOUT ANY WARRANTY; without even the implied warranty of
!*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!*   GNU General Public License for more details.
!*
!*   You should have received a copy of the GNU General Public License
!*   along with this program; if not, write to the Free Software
!*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!*
!********************************************************************************
module snf_vroa
  use f90_kind
  use snf_error

  private
  public :: create_aten_dxyz,print_aten_dxyz
  public :: create_aten_dq,create_aten_dq_intonly,calcvroa
  public :: expand_gten,symm_gten

  !************************************************************************

contains

  !************************************************************************

  subroutine create_aten_dxyz(polar_dxyz,pols,distortion)
    implicit none
    !----------------------------------------------------------------
    ! calculation of first derivative of a tensor with arbitrary number
    ! of tensor components according to 
    ! the following 3-point formula:
    ! 
    !  d alpha_ij     pols_ij(xi+e(xi)) - pols_ij(xi-e(xi))  
    ! ------------ = ---------------------------------------   
    !    dxi                     2*e(xi) 
    !
    ! a 5- and 7-point central difference formulae are also implemented
    !           
    ! Note: the units are atomic units and the step size is therefore
    !       required in bohr!
    !
    ! written by M. Reiher, Uni Erlangen, June 2003
    !----------------------------------------------------------------
    real(double),intent(inout) :: polar_dxyz(:,:)
    real(double),intent(in) :: pols(:,:,:)
    real(double),intent(in) :: distortion(:)
    integer :: idum,i
    polar_dxyz=0.0d0
    idum=ubound(pols,3)
    if (idum.eq.2) then
      do i=1,ubound(polar_dxyz,1)
        polar_dxyz(i,:) = 0.5d0/distortion(:)*(pols(i,:,1)-pols(i,:,2))
      end do
    else if (idum.eq.4) then   
      do i=1,ubound(polar_dxyz,1)
        polar_dxyz(i,:) = 1.0d0/(24.0d0*distortion(:))*(-2.0d0*pols(i,:,1)&
          &+16.0d0*pols(i,:,2)-16.0d0*pols(i,:,3)+2.0d0*pols(i,:,4))
      end do
    else if (idum.eq.6) then   
      do i=1,ubound(polar_dxyz,1)
        polar_dxyz(i,:) = 1.0d0/(720.0d0*distortion(:))*(-12.0d0*pols(i,:,6)&
         & +108.0d0*pols(i,:,5)-540.0d0*pols(i,:,4)+540.0d0*pols(i,:,3) &
         & -108.0d0*pols(i,:,2)+12.0d0*pols(i,:,1))
      end do
    else
       call serror('create_aten_dxyz',2,'Wrong ctrl%numderiv value!') 
    end if
  end subroutine create_aten_dxyz
  !************************************************************************


  subroutine create_aten_dq_intonly(polar_dqi,polar,nummodes,normmodes,cstep,numderiv,natoms,molecule)
    use snf_molecule
    implicit none
    !----------------------------------------------------------------
    ! calculation of first derivative of A-tensor with
    ! respect to mass-weighted normal coordinates using the
    ! following formula:
    !
    !  d polar    polar(qi+e(qi) - polar(qi-e(qi))
    ! -------- = ---------------------------------
    !    dqi                 2*distortion
    !
    ! we use one distortion value for all normal modes
    ! Note: the units are atomic units and the step size is therefore
    !       required in bohr!
    !
    ! written by C. Herrmann, Uni Bonn 2005,
    ! based on create_polar_dxyz, written by Carsten Kind and
    ! Markus Reiher, J. Neugebauer, Uni Erlangen, 1999-2002
    !----------------------------------------------------------------
    real(double),intent(inout):: polar_dqi(:,:)
    real(double),intent(in):: polar(:,:,:)
    real(double) :: cstep
    integer :: i,j ,natoms,nummodes,numderiv
    real(double) :: normmodes(natoms*3,natoms*3)
    real(double) :: zthres, sum, red_mass(nummodes)
    type(type_atom),pointer :: molecule(:)
    red_mass(:) = 0.d0
    do i = 1,ubound(polar,1)
    if (numderiv.eq.1) then
       polar_dqi(i,:) = 0.5d0/cstep*(polar(i,:,1)-polar(i,:,2))
    else if (numderiv.eq.2) then
       polar_dqi(i,:) = 1.0d0/(24.0d0*cstep)*(-2.0d0*polar(i,:,1)&
        & +16.0d0*polar(i,:,2)-16.0d0*polar(i,:,3)+2.0d0*polar(i,:,4))
    else if (numderiv.eq.3) then
       polar_dqi(i,:) = 1.0d0/(720.0d0*cstep)*(-12.0d0*polar(i,:,6)&
       & +108.0d0*polar(i,:,5)-540.0d0*polar(i,:,4)+540.0d0*polar(i,:,3) &
       & -108.0d0*polar(i,:,2)+12.0d0*polar(i,:,1))
    else
       call serror('create_aten_dq_intonly',2,'Wrong ctrl%numderiv value!')
    end if
    end do

! calculate reduced mass

          do i = 1,nummodes
          do j=1,ubound(molecule,1)*3
              red_mass(i) = red_mass(i) + &
                   & (normmodes(i,j)**2 &
                   & /molecule((j+2)/3)%mass)
           enddo
           red_mass(i)=1.d0/red_mass(i)
! divide derivative by reduced mass
           polar_dqi(:,i) = polar_dqi(:,i)/dsqrt(red_mass(i))
           end do


  end subroutine create_aten_dq_intonly


  !************************************************************************

  subroutine print_aten_dxyz(polar_dxyz,filename)
    implicit none
    !----------------------------------------------------------------
    ! output routine
    !
    ! written by Markus Reiher, Uni Erlangen, June 2003
    !----------------------------------------------------------------
    character*(*) :: filename
    integer :: i,k
    real(double),intent(in) :: polar_dxyz(:,:)

    open(unit=45,file=filename,status='unknown')
    write(45,*) 'cartesian tensor derivatives'
    write(45,'(3d18.10)') ((polar_dxyz(i,k),i=1,ubound(polar_dxyz,1)),k=1,ubound(polar_dxyz,2))
    close (45)
  end subroutine print_aten_dxyz

  !************************************************************************

  subroutine create_aten_dq(polar_dxyz,norm_modes,molecule,polar_dq)
    !----------------------------------------------------------------
    ! written by M. Reiher, Uni Erlangen, June 2003
    !
    ! Note: Cartesian tensor derivatives are given in a.u.
    !----------------------------------------------------------------
    use snf_molecule, only: type_atom
    implicit none
    real(double),intent(in) :: polar_dxyz(:,:),norm_modes(:,:)
    type(type_atom), pointer :: molecule(:)
    real(double),intent(inout):: polar_dq(:,:)
    integer :: i,j,k,nvar,irrep_dim

    nvar=ubound(norm_modes,1)
    irrep_dim=ubound(norm_modes,2)
    polar_dq=0.0d0
    do k=1,ubound(polar_dxyz,1)
       do i=1,nvar
          do j=1,irrep_dim
             polar_dq(k,j)= polar_dq(k,j) + polar_dxyz(k,i)*norm_modes(i,j) &
                  &        / dsqrt(molecule((i+2)/3)%mass)
          enddo
       enddo
    enddo
  end subroutine create_aten_dq

  !************************************************************************

  subroutine expand_gten(gten_dxyz,representation,symm_index,char_table)
    use snf_char_table, only : type_char_table
    use snf_symmat,only : symel_mat
    !----------------------------------------------------------------
    ! copy second-rank 3x3 tensor derivatives calculated 
    ! on irreducible atoms to the
    ! reducible atoms. this is done by multiplying symmetry representations
    ! in the pol basis from the left and symmetry representations in the
    ! derivate basis from the right with the tensor component derivatives
    ! and selecting only the required columns.
    !
    ! written by Markus Reiher, Uni Erlangen, June 2003
    !----------------------------------------------------------------
    implicit none
    type(type_char_table), intent(in) :: char_table
    real(double),intent(inout) :: gten_dxyz(:,:)
    real(double),pointer :: rot_polar(:,:),expanded_polar(:,:),polar_tmp(:,:)
    real(double),intent(in) :: representation(:,:,:)
    integer, intent(in) :: symm_index(:)
    integer :: s,i,offset,nvar,counter,j,k,natoms
    real(double) :: pol(3,3),transpol(3,3)
    nvar=ubound(gten_dxyz,2)
    natoms=nvar/3
    allocate(rot_polar(9,nvar),polar_tmp(9,nvar),expanded_polar(9,nvar))
    do s=1,char_table%nsymel
       do i=1,nvar
          ! copy part of long packed matrix to square matrix
          counter=0
          do j=1,3
             do k=j,3
                counter=counter+1
                pol(k,j) = gten_dxyz(counter,i)
             enddo
          enddo
          ! rotate square matrix
          transpol = matmul(symel_mat(char_table%symel(s)),&
               &            matmul(pol,transpose(symel_mat(char_table%symel(s)))))
          ! copy back to packed long matrix
          counter=0
          do j=1,3
             do k=j,3
                counter=counter+1
                polar_tmp(counter,i)=transpol(k,j)
             enddo
          enddo
       enddo

       ! perform transformation of tensor gradient
       rot_polar=matmul(polar_tmp,representation(:,:,s))
       ! (9 x nvar ) = (9 x nvar) * (nvar x nvar)
       ! move selected data per atom to final place
       do i=1,natoms
          if(symm_index(i) .eq. s) then
             offset=(i-1)*3
             expanded_polar(:,offset+1:offset+3)=rot_polar(:,offset+1:offset+3)
             !              9,3                            9,3
          endif
       enddo
    enddo
    do i=1,9
      do j=1,nvar
        gten_dxyz(i,j) = expanded_polar(i,j)
      end do
    end do
    deallocate(rot_polar,polar_tmp,expanded_polar)
  end subroutine expand_gten

  !************************************************************************

  subroutine symm_gten(gten_dxyz,rep,char_table)
    use snf_char_table, only : type_char_table
    use snf_symmat,only : symel_mat
    !----------------------------------------------------------------
    ! this subroutine transforms the  
    ! second-rank 3x3 tensor component derivatives
    ! according to all symmetry operations of the point group.
    ! all resulting matrices are added and then diveded by the
    ! number of symmetry operations. a symmetry check is performed.
    !
    ! written by Markus Reiher, Uni Erlangen, June 2003
    !----------------------------------------------------------------
    implicit none
    type(type_char_table), intent(in) :: char_table
    real(double),intent(inout) :: gten_dxyz(:,:)
    real(double),pointer :: rot_polar(:,:),polar_tmp(:,:),tmpmat(:,:)
    real(double),intent(in) :: rep(:,:,:)
    real(double) :: errsum
    integer :: s,i,nvar,counter,j,k
    character(len=60) :: errbuff
    real(double) :: pol(3,3),transpol(3,3)
    nvar=ubound(gten_dxyz,2)
    allocate(rot_polar(9,nvar),polar_tmp(9,nvar),tmpmat(9,nvar))
    rot_polar=0.0d0
    do s=1,char_table%nsymel
      do i=1,nvar
        ! copy part of long packed matrix to square matrix
        counter=0
        do j=1,3
           do k=j,3
              counter=counter+1
              pol(k,j) = gten_dxyz(counter,i)
           enddo
        enddo
        ! rotate square matrix
        transpol = matmul(symel_mat(char_table%symel(s)),&
             &            matmul(pol,transpose(symel_mat(char_table%symel(s)))))
        ! copy back to packed long matrix
        counter=0
        do j=1,3
           do k=j,3
              counter=counter+1
              polar_tmp(counter,i)=transpol(k,j)
           enddo
        enddo
      enddo
      ! perform transformation of gradient
      tmpmat = matmul(polar_tmp,rep(:,:,s))
      rot_polar=rot_polar + tmpmat
      do i=1,9
        do j=1,nvar
          errsum=errsum+abs(tmpmat(i,j)-gten_dxyz(i,j))
        end do
      enddo
      errsum=errsum/dble(nvar)/9.0d0
      if(errsum.gt.1.0d-3) then
        write(errbuff,'(a,f8.4,a,i3)') 'average error = ',errsum, ' for s = ',s
        call serror('symm_gten',0,trim(adjustl(errbuff)))
      end if
    enddo
    rot_polar=rot_polar/dble(char_table%nsymel)
    do i=1,9
      do j=1,nvar
        gten_dxyz(i,j) = rot_polar(i,j)
      end do
    end do
    deallocate(rot_polar,polar_tmp,tmpmat)
  end subroutine symm_gten

  !************************************************************************

  subroutine calcvroa(freqs,polar9_dq,gtenao_dq,gtenlao_dq,aten_dq,&
                      &NCORD,NUMMOD,NFRVAL,MXFR,FRVAL,LUPRI,Atensor_0,&
                      &progno)
    implicit none
    !----------------------------------------------------------------
    ! driver for DALTON VROA evaluation routines
    !
    ! written by Markus Reiher, Uni Erlangen, June 2003
    ! (a bit) modified by S. Luber, ETH Zuerich, October 2006
    !----------------------------------------------------------------

    real(double),intent(in) :: freqs(:),polar9_dq(:,:,:),gtenao_dq(:,:,:),gtenlao_dq(:,:,:),aten_dq(:,:,:)
    real(double),intent(in) :: FRVAL(:)
    real(double),pointer    :: ROAAFQ(:,:,:,:),ROAGNQ(:,:,:,:),ROAGLQ(:,:,:,:),ROAAQ(:,:,:,:,:)
    real(double),pointer    ::  dumROAAQ(:,:,:,:,:)
    real(double),pointer    :: BAL(:),ALMN(:),CONVER(:),BOLTZ(:)
    real(double),pointer    :: BGML(:),BGMNL(:),BA(:),GMMNL(:),GMMNNL(:),RMINN (:),DPOLN (:),DZL(:),&
                              &DZNL(:),DXL(:),DXNL(:),D0L(:),D0NL(:),DBL(:),DBNL(:),CZL(:),CZNL(:),&
                              &CXL(:),CXNL(:),C0L(:),C0NL(:),CBL(:),CBNL(:),FREQ(:)
    integer NFRVAL,NUMMOD,MXFR,LUPRI,NCORD,i,j,idum,k,l,idum2,XTKAYS,progno
!SL
    logical Atensor_0
    parameter(XTKAYS=219474.6313710D0)
!... store derivatives in format of the DALTON routine
    allocate(FREQ(NUMMOD))
    allocate(ROAAFQ(3,3,NFRVAL,NCORD))
    allocate(ROAGNQ(3,3,NFRVAL,NCORD))
    allocate(ROAGLQ(3,3,NFRVAL,NCORD))
    allocate(ROAAQ(3,3,3,NFRVAL,NCORD))
    allocate(dumROAAQ(3,3,3,NFRVAL,NCORD))
    FREQ(:)=freqs(:)
    do i=1,NUMMOD
      FREQ(i)=FREQ(i)/XTKAYS
    end do
!   ... sort tensors
    idum=0
    do i=1,3
      do j=1,3
        idum=idum+1
        do k=1,NFRVAL
          do l=1,NCORD
            ROAAFQ(i,j,k,l)=polar9_dq(idum,l,k)
            ROAGNQ(i,j,k,l)=gtenao_dq(idum,l,k)
            ROAGLQ(i,j,k,l)=gtenlao_dq(idum,l,k)
          end do
        end do
      end do
    end do 
!   ... first 12 entries of A
    idum=0
    do j=1,3
      do i=1,3
        idum=idum+1
        do k=1,NFRVAL
          do l=1,NCORD
            ROAAQ(1,i,j,k,l)=aten_dq(idum,l,k) 
          end do
        end do
      end do
      idum=idum+1
      do k=1,NFRVAL
        do l=1,NCORD
          ROAAQ(2,1,j,k,l)=aten_dq(idum,l,k)  
        end do
      end do
    end do
!   ... second 12 entries of A
    do j=1,3
      do i=1,2
        idum=idum+1
        if (i.eq.1) idum2=2
        if (i.eq.2) idum2=3
        do k=1,NFRVAL
          do l=1,NCORD
            ROAAQ(2,idum2,j,k,l)=aten_dq(idum,l,k)
          end do
        end do
      end do
      do i=1,2
        idum=idum+1
        do k=1,NFRVAL
          do l=1,NCORD
            ROAAQ(3,i,j,k,l)=aten_dq(idum,l,k)
          end do
        end do
      end do
    end do
!   ... last 3 entries of A
    do j=1,3
      idum=idum+1
      do k=1,NFRVAL
        do l=1,NCORD
          ROAAQ(3,3,j,k,l)=aten_dq(idum,l,k)
        end do
      end do
    end do
    if (idum.ne.27) stop " MIRACLE !!!"
! Reordering is necessary, but how can this be?
    dumROAAQ(:,:,:,:,:)=ROAAQ(:,:,:,:,:)
    do i=1,3
      do j=1,3
        ROAAQ(i,:,j,:,:)=dumROAAQ(j,:,i,:,:)
      end do
    end do
!
!... provide scratch arrays for DALTON routine
    allocate(BAL(NCORD))
    allocate(ALMN(NCORD))
    allocate(CONVER(NCORD))
    allocate(BOLTZ(NCORD))
    allocate(BGML(NCORD))
    allocate(BGMNL(NCORD))
    allocate(BA(NCORD))
    allocate(GMMNL(NCORD))
    allocate(GMMNNL(NCORD))
    allocate(RMINN(NCORD))
    allocate(DPOLN(NCORD))
    allocate(DZL(NCORD))
    allocate(DZNL(NCORD))
    allocate(DXL(NCORD))
    allocate(DXNL(NCORD))
    allocate(D0L(NCORD))
    allocate(D0NL(NCORD))
    allocate(DBL(NCORD))
    allocate(DBNL(NCORD))
    allocate(CZL(NCORD))
    allocate(CZNL(NCORD))
    allocate(CXL(NCORD))
    allocate(CXNL(NCORD))
    allocate(C0L(NCORD))
    allocate(C0NL(NCORD))
    allocate(CBL(NCORD))
    allocate(CBNL(NCORD))
!
    call VIBRO1_MOD(NFRVAL,MXFR,FREQ,ROAAFQ,ROAGNQ,ROAGLQ,ROAAQ,&
     &              BAL,ALMN,CONVER,BOLTZ,NCORD,NUMMOD,FRVAL,LUPRI,&
     &              BGML,BGMNL,BA,DZL,DZNL,&
     &              DXL,DXNL,D0L,D0NL,DBL,DBNL,&
     &              GMMNL,GMMNNL,RMINN,DPOLN,CZL,CZNL,CXL,CXNL,&
     &              C0L,C0NL,CBL,CBNL,Atensor_0,progno)
!
    deallocate(ROAAFQ)
    deallocate(ROAGNQ)
    deallocate(ROAGLQ)
    deallocate(ROAAQ)
    deallocate(BGML)
    deallocate(BGMNL)
    deallocate(BA)
    deallocate(GMMNL)
    deallocate(GMMNNL)
    deallocate(RMINN)
    deallocate(DPOLN)
    deallocate(DZL)   
    deallocate(DZNL) 
    deallocate(DXL) 
    deallocate(DXNL)
    deallocate(D0L) 
    deallocate(D0NL)
    deallocate(DBL)
    deallocate(DBNL)
    deallocate(CZL)
    deallocate(CZNL) 
    deallocate(CXL)
    deallocate(CXNL)
    deallocate(C0L)
    deallocate(C0NL)
    deallocate(CBL)
    deallocate(CBNL)
  end subroutine calcvroa

  !************************************************************************

end module snf_vroa
